@@ -30,6 +30,7 @@
 #include <boost/algorithm/string.hpp>
 #include <boost/algorithm/string/classification.hpp>
 #include <boost/algorithm/string/replace.hpp>
+#include <boost/format.hpp>
 
 #include <iostream>
 #include "base32.hh"
@@ -79,7 +80,7 @@ void RecordTextReader::xfr64BitInt(uint64_t &val)
 
   size_t pos;
   val=std::stoull(d_string.substr(d_pos), &pos);
-  
+
   d_pos += pos;
 }
 
@@ -92,8 +93,8 @@ void RecordTextReader::xfr32BitInt(uint32_t &val)
     throw RecordTextException("expected digits at position "+std::to_string(d_pos)+" in '"+d_string+"'");
 
   size_t pos;
-  val=pdns_stou(d_string.c_str()+d_pos, &pos);
- 
+  val = pdns::checked_stoi<uint32_t>(d_string.c_str() + d_pos, &pos);
+
   d_pos += pos;
 }
 
@@ -101,7 +102,7 @@ void RecordTextReader::xfrTime(uint32_t &val)
 {
   struct tm tm;
   memset(&tm, 0, sizeof(tm));
-  
+
   uint64_t itmp;
   xfr64BitInt(itmp);
 
@@ -123,7 +124,7 @@ void RecordTextReader::xfrTime(uint32_t &val)
 
   tm.tm_year-=1900;
   tm.tm_mon-=1;
-  val=(uint32_t)Utility::timegm(&tm); 
+  val=(uint32_t)Utility::timegm(&tm);
 }
 
 void RecordTextReader::xfrIP(uint32_t &val)
@@ -157,7 +158,7 @@ void RecordTextReader::xfrIP(uint32_t &val)
       if(octet > 255)
         throw RecordTextException("unable to parse IP address");
     }
-    else if(dns_isspace(d_string.at(d_pos)) || d_string.at(d_pos) == ',') 
+    else if(dns_isspace(d_string.at(d_pos)) || d_string.at(d_pos) == ',')
       break;
     else {
       throw RecordTextException(string("unable to parse IP address, strange character: ")+d_string.at(d_pos));
@@ -181,10 +182,10 @@ void RecordTextReader::xfrIP6(std::string &val)
   struct in6_addr tmpbuf;
 
   skipSpaces();
-  
+
   size_t len;
   // lookup end of value - think of ::ffff encoding too, has dots in it!
-  for(len=0; 
+  for(len=0;
       d_pos+len < d_string.length() && (isxdigit(d_string.at(d_pos+len)) || d_string.at(d_pos+len) == ':' || d_string.at(d_pos+len)=='.');
     len++);
 
@@ -193,7 +194,7 @@ void RecordTextReader::xfrIP6(std::string &val)
 
   // end of value is here, try parse as IPv6
   string address=d_string.substr(d_pos, len);
-  
+
   if (inet_pton(AF_INET6, address.c_str(), &tmpbuf) != 1) {
     throw RecordTextException("while parsing IPv6 address: '" + address + "' is invalid");
   }
@@ -248,7 +249,7 @@ void RecordTextReader::xfr8BitInt(uint8_t &val)
     throw RecordTextException("Overflow reading 8 bit integer from record content"); // fixme improve
 }
 
-// this code should leave all the escapes around 
+// this code should leave all the escapes around
 void RecordTextReader::xfrName(DNSName& val, bool, bool)
 {
   skipSpaces();
@@ -259,7 +260,7 @@ void RecordTextReader::xfrName(DNSName& val, bool, bool)
   while(d_pos < d_end) {
     if(strptr[d_pos]!='\r' && dns_isspace(strptr[d_pos]))
       break;
-      
+
     d_pos++;
   }
   sval = DNSName(std::string(strptr+begin_pos, strptr+d_pos));
@@ -277,9 +278,9 @@ static bool isbase64(char c, bool acceptspace)
     return acceptspace;
   if(c >= '0' && c <= '9')
     return true;
-  if(c >= 'a' && c <= 'z') 
+  if(c >= 'a' && c <= 'z')
     return true;
-  if(c >= 'A' && c <= 'Z') 
+  if(c >= 'A' && c <= 'Z')
     return true;
   if(c=='+' || c=='/' || c=='=')
     return true;
@@ -290,7 +291,7 @@ void RecordTextReader::xfrBlobNoSpaces(string& val, int len) {
   skipSpaces();
   int pos=(int)d_pos;
   const char* strptr=d_string.c_str();
-  while(d_pos < d_end && isbase64(strptr[d_pos], false)) 
+  while(d_pos < d_end && isbase64(strptr[d_pos], false))
     d_pos++;
 
   string tmp;
@@ -298,7 +299,7 @@ void RecordTextReader::xfrBlobNoSpaces(string& val, int len) {
   boost::erase_all(tmp," ");
   val.clear();
   B64Decode(tmp, val);
-  
+
   if (len>-1 && val.size() != static_cast<size_t>(len))
     throw RecordTextException("Record length "+std::to_string(val.size()) + " does not match expected length '"+std::to_string(len));
 }
@@ -310,7 +311,7 @@ void RecordTextReader::xfrBlob(string& val, int)
   const char* strptr=d_string.c_str();
   while(d_pos < d_end && isbase64(strptr[d_pos], true))
     d_pos++;
-  
+
   string tmp;
   tmp.assign(d_string.c_str()+pos, d_string.c_str() + d_pos);
   boost::erase_all(tmp," ");
@@ -526,7 +527,7 @@ static void HEXDecode(const char* begin, const char* end, string& out)
       val = 16*hextodec(*begin);
       mode=1;
     } else {
-      val += hextodec(*begin); 
+      val += hextodec(*begin);
       out.append(1, (char) val);
       mode = 0;
       val = 0;
@@ -731,12 +732,12 @@ void RecordTextWriter::xfrIP6(const std::string& val)
 
   if(!d_string.empty())
    d_string.append(1,' ');
-  
+
   val.copy(tmpbuf,16);
 
   if (inet_ntop(AF_INET6, tmpbuf, addrbuf, sizeof addrbuf) == nullptr)
     throw RecordTextException("Unable to convert to ipv6 address");
-  
+
   d_string += std::string(addrbuf);
 }
 
@@ -759,7 +760,7 @@ void RecordTextWriter::xfrTime(const uint32_t& val)
 {
   if(!d_string.empty())
     d_string.append(1,' ');
-  
+
   struct tm tm;
   time_t time=val; // Y2038 bug!
   gmtime_r(&time, &tm);
@@ -784,7 +785,7 @@ void RecordTextWriter::xfrName(const DNSName& val, bool, bool noDot)
 {
   if(!d_string.empty())
     d_string.append(1,' ');
-  
+
   if(d_nodot) {
     d_string+=val.toStringRootDot();
   }
@@ -961,7 +962,7 @@ int main(int argc, char**argv)
 try
 {
   RecordTextReader rtr(argv[1], argv[2]);
-  
+
   unsigned int order, pref;
   string flags, services, regexp, replacement;
   string mx;
@@ -987,7 +988,7 @@ try
   rtw.xfrName(replacement);
 
   cout<<"Regenerated: '"<<out<<"'\n";
-  
+
 }
 catch(std::exception& e)
 {
